import {
	GetServerSideProps,
	NextApiHandler,
	NextApiRequest as DefaultNextApiRequest,
	NextApiResponse as DefaultNextApiResponse,
	PageConfig,
} from 'next';
import * as nookies from 'nookies';
import { IncomingMessage } from 'http';
import busboy from 'busboy';
import { deserialize, serialize } from 'seroval';
import {
	NextApiResponse,
	NextApiRequest,
	ENCTYPE_APPLICATION_JSON,
	ENCTYPE_X_WWW_FORM_URLENCODED,
	ENCTYPE_MULTIPART_FORM_DATA, ENCTYPE_APPLICATION_OCTET_STREAM,
} from './common';

let BODY_COOKIE_KEY: string;
let STATUS_CODE_COOKIE_KEY: string;
let STATUS_MESSAGE_COOKIE_KEY: string;
let CONTENT_TYPE_COOKIE_KEY: string;

const generateContentTypeCookieKey = () => {
	CONTENT_TYPE_COOKIE_KEY = `ifct${Date.now()}`;
};

const generateBodyCookieKey = () => {
	BODY_COOKIE_KEY = `ifb${Date.now()}`;
};

const generateStatusCodeCookieKey = () => {
	STATUS_CODE_COOKIE_KEY = `ifsc${Date.now()}`;
};

const generateStatusMessageCookieKey = () => {
	STATUS_MESSAGE_COOKIE_KEY = `ifsm${Date.now()}`;
};

const getBody = (req: IncomingMessage) => new Promise<Buffer>((resolve, reject) => {
	let body = Buffer.from('');

	req.on('data', (chunk) => {
		body = Buffer.concat([body, chunk]);
	});

	req.on('error', (err) => {
		reject(err);
	});

	req.on('end', () => {
		resolve(body);
	});
});

export namespace destination {
	export const getServerSideProps = (
		gspFn?: GetServerSideProps,
	): GetServerSideProps => async (ctx) => {
		const req: NextApiRequest = {
			query: ctx.query,
			body: null,
		};
		const { method = 'GET' } = ctx.req;

		if (!['GET', 'HEAD'].includes(method.toUpperCase())) {
			const body = await getBody(ctx.req);
			req.body = body.toString('utf-8');
		}

		const cookies = nookies.parseCookies(ctx);
		const res: NextApiResponse = {};

		// TODO how to properly remove cookies without leftovers?
		if (STATUS_CODE_COOKIE_KEY in cookies) {
			ctx.res.statusCode = Number(cookies[STATUS_CODE_COOKIE_KEY] || '200');
			nookies.destroyCookie(ctx, STATUS_CODE_COOKIE_KEY, {
				path: '/',
				httpOnly: true,
			});
		}

		if (STATUS_MESSAGE_COOKIE_KEY in cookies) {
			ctx.res.statusMessage = cookies[STATUS_MESSAGE_COOKIE_KEY] || '';
			nookies.destroyCookie(ctx, STATUS_MESSAGE_COOKIE_KEY, {
				path: '/',
				httpOnly: true,
			});
		}

		if (BODY_COOKIE_KEY in cookies) {
			const resBody = cookies[BODY_COOKIE_KEY];
			if (cookies[CONTENT_TYPE_COOKIE_KEY] === ENCTYPE_APPLICATION_JSON) {
				res.body = deserialize(resBody);
			} else if (cookies[CONTENT_TYPE_COOKIE_KEY] === ENCTYPE_APPLICATION_OCTET_STREAM) {
				res.body = deserialize(resBody);
			} else {
				const c = console;
				c.warn('Could not parse response body, returning nothing');
			}
			nookies.destroyCookie(ctx, BODY_COOKIE_KEY, {
				path: '/',
				httpOnly: true,
			});
			nookies.destroyCookie(ctx, CONTENT_TYPE_COOKIE_KEY, {
				path: '/',
				httpOnly: true,
			});
		}

		let gspResult;
		if (gspFn) {
			gspResult = await gspFn(ctx);
		} else {
			gspResult = {
				props: {},
			};
		}
		if ('props' in gspResult) {
			return {
				...gspResult,
				props: {
					...gspResult.props,
					req,
					res,
				},
			};
		}

		// redirect/not found will be treated as default behavior
		return gspResult;
	};
}

export namespace action {
	const parseMultipartFormData = async (
		req: IncomingMessage,
	) => new Promise<Record<string, unknown>>((resolve, reject) => {
		const body: Record<string, unknown> = {};
		const bb = busboy({
			headers: req.headers,
		});

		bb.on('file', (name, file, info) => {
			const {
				filename,
				mimeType: mimetype,
			} = info;

			let fileData = Buffer.from('');

			file.on('data', (data) => {
				fileData = Buffer.concat([fileData, data]);
			});

			file.on('close', () => {
				body[name] = new File([fileData.buffer], filename, {
					type: mimetype,
				});
			});
		});

		bb.on('field', (name, value) => {
			body[name] = value;
		});

		bb.on('close', () => {
			resolve(body);
		});

		bb.on('error', (error) => {
			reject(error);
		});

		req.pipe(bb);
	});

	const deserializeBody = async (req: IncomingMessage) => {
		const contentType = req.headers['content-type'];
		const encoding = (req.headers['content-encoding'] ?? 'utf-8') as BufferEncoding;
		if (contentType === ENCTYPE_APPLICATION_JSON) {
			const bodyRaw = await getBody(req);
			return JSON.parse(bodyRaw.toString(encoding)) as Record<string, unknown>;
		}
		if (contentType === ENCTYPE_X_WWW_FORM_URLENCODED) {
			const bodyRaw = await getBody(req);
			return Object.fromEntries(
				new URLSearchParams(bodyRaw.toString(encoding)).entries(),
			);
		}
		if (contentType?.startsWith(`${ENCTYPE_MULTIPART_FORM_DATA};`)) {
			return parseMultipartFormData(req);
		}
		const bodyRaw = await getBody(req);
		return bodyRaw.toString('binary');
	};

	export const getApiConfig = (customConfig = {} as PageConfig) => ({
		api: {
			...(customConfig.api ?? {}),
			bodyParser: false,
		},
	});

	export const wrapApiHandler = (fn: NextApiHandler): NextApiHandler => async (req, res) => {
		const body = await deserializeBody(req);
		const reqMut = req as unknown as Record<string, unknown>;
		reqMut.body = body;
		return fn(reqMut as unknown as DefaultNextApiRequest, res);
	};

	export const getServerSideProps = (fn: NextApiHandler): GetServerSideProps => async (ctx) => {
		const { referer } = ctx.req.headers;

		const mockReq = {
			...ctx.req,
			body: await deserializeBody(ctx.req),
		} as DefaultNextApiRequest;

		let data: unknown = null;
		let contentType: string | undefined;
		const mockRes = {
			// todo handle other nextapiresponse methods (e.g. setting headers, writeHead, etc.)
			statusMessage: '',
			statusCode: 200,
			status(code: number) {
				// should we mask error status code to Bad Gateway?
				this.statusCode = code;
				return mockRes;
			},
			send: (raw?: unknown) => {
				// todo: how to transfer binary response in a more compact way?
				if (typeof raw === 'undefined' || raw === null) {
					return;
				}

				if (raw instanceof Buffer) {
					contentType = ENCTYPE_APPLICATION_OCTET_STREAM;
				}

				data = serialize(raw);
			},
			json: (raw: unknown) => {
				contentType = ENCTYPE_APPLICATION_JSON;
				data = serialize(raw);
			},
		} as DefaultNextApiResponse;

		await fn(mockReq, mockRes);

		generateStatusCodeCookieKey();
		nookies.setCookie(ctx, STATUS_CODE_COOKIE_KEY, mockRes.statusCode.toString(), {
			maxAge: 30 * 24 * 60 * 60,
			path: '/',
			httpOnly: true,
		});

		generateStatusMessageCookieKey();
		nookies.setCookie(ctx, STATUS_MESSAGE_COOKIE_KEY, mockRes.statusMessage, {
			maxAge: 30 * 24 * 60 * 60,
			path: '/',
			httpOnly: true,
		});

		if (data) {
			generateBodyCookieKey();
			nookies.setCookie(ctx, BODY_COOKIE_KEY, data as string, {
				maxAge: 30 * 24 * 60 * 60,
				path: '/',
				httpOnly: true,
			});

			if (contentType) {
				generateContentTypeCookieKey();
				nookies.setCookie(ctx, CONTENT_TYPE_COOKIE_KEY, contentType, {
					maxAge: 30 * 24 * 60 * 60,
					path: '/',
					httpOnly: true,
				});
			}
		}

		return {
			redirect: {
				destination: referer,
				statusCode: 307,
			},
			props: {
				query: ctx.query,
				body: data,
			},
		};
	};
}
